<?php
namespace SupplyTek\Controllers;

use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DTO\Response;
use SupplyTek\Utilities\CrudSupplyTek;
use SupplyTek\Utilities\SessionHandler;
use SupplyTek\Utilities\FormatData;

/**
 * Searches for user information and unit data logged into the system.
 *
 * @author Grazielle Ferreira Antunes Brum
 */
class UserController 
{
    protected $crud;
    protected $session;
    protected $format;

    /**
     * Class constructor
     * 
     * Initialize an object's properties upon creation of the object.
     * 
     * @param CrudSupplyTek $crudSupplyTek Object that allows to perform queries from datasource
     * @param SessionHandler $sessionHandler Object that allows to manipulate session
     * @param FormatData $formatData Object that allows to format data
     * 
     * @return void
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function __construct (CrudSupplyTek $crudSupplyTek, SessionHandler $sessionHandler, FormatData $formatData) 
    {
        $this->crud = $crudSupplyTek;
        $this->session = $sessionHandler;
        $this->format = $formatData;
    }
    
    /**
     * Return all approvers from unit
     * 
     * @param Request $request Object that receives data from the caller
     * @param Response $response Object that sends back data to the caller
     *
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    public function getApprovers (Request $request, Response $response)
    {
        try {
            $params["UNIT_ID"] = $this->session->getSessionData("CDFILIAL");
            $params["ORG_ID"] = $this->session->getSessionData("NRORG");
            
            $approvers = $this->crud->find($request, $params);
            
            $response->setMetadata($approvers);
        } catch (\Exception $e) {
            throw new \Exception($e->getMessage(), 500);
        }
    }

    /**
     * Prepares search data for the service and returns user and unit and billing information to the front end.
     *
     * @param Request $request Object that receives data from the caller.
     * @param Response $response Object that sends back data to the caller.
     *
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Grazielle Ferreira Antunes Brum
     */
    public function getUserInfo(Request $request, Response $response)
    {
        try {
            $profile = $this->session->getSessionData("IDPERFILACES");
            $operatorId = $this->session->getSessionData("CDOPERADOR");
            $unitId = $this->session->getSessionData("CDFILIAL");
            $orgId = $this->session->getSessionData("NRORG");
            
            switch ($profile) {
                case "GO":
                    $params["OPERATOR_ID"] = $operatorId;
                    $params["UNIT_ID"] = $unitId;
                    $params["ORG_ID"] = $orgId;
                    
                    $userInfo = $this->crud->executeQuery("user-info-GO", $params);
                    
                    $array = $this->getUserInfoGO($userInfo);
                break;
                
                case "GU":
                    $params["UNIT_ID"] = $unitId;
                    $billingInfo = $this->crud->executeQuery("billing-info", $params);
                    
                    $params["ORG_ID"] = $orgId;
                    $approversInfo = $this->crud->executeQuery("approvers", $params);
                    
                    $params["OPERATOR_ID"] = $operatorId;
                    $userInfo = $this->crud->executeQuery("user-info", $params);
                    
                    $array = $this->getUserInfoGU($billingInfo, $approversInfo, $userInfo);
                break;
            }
            
            $response->setMetadata($array);
        } catch (\Exception $e) {
            throw $e;
        }
    }
            
    /**
     * Create and return necessary data for GU User Info
     *
     * Receives the data needed to create the array with all the info
     * needed to return for the GU User Info process.
     * 
     * @param Array $billingInfo Array with billing information.
     * @param Array $approversInfo Array with approvers information.
     * @param Array $userInfo Array with user information.
     *
     * @return array
     * @throws Exception
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function getUserInfoGU($billingInfo, $approversInfo, $userInfo)
    {
        try {
            if (isset($billingInfo[0])) {
                $billing = $billingInfo[0];
                $addressBillingMandate = $this->format->formatAddressBillingMandate($billing);
            } else {
                $addressBillingMandate["BILLING_ADDRESS"] = "";
                $addressBillingMandate["MANDATE_ADDRESS"] = "";
            }
            
            if (isset($approversInfo[0])) {
                $approvers = $this->returnApproversGU($approversInfo);
            } else {
                $approvers = [];
            }
            
            if (isset($userInfo[0])) {
                $user = $userInfo[0];
                
                $address = $this->format->formatAddress($user["STREET"], $user["NEIGHBORHOOD"], $user["COUNTY"], $user["STATE"], $user["POSTAL_CODE"]);
                $applicant = $this->format->capitalizeWords($user["OPERATOR_NAME"]);
                $phone = $this->format->formatPhone($user["PHONE"], true);
            } else {
                $address = "";
                $applicant = "";
                $phone = "";
            }
            
            $array = array(
                "APPLICANT" => $applicant,
                "NAME" => $billing["NMFILIAL"],
                "PHONE" => $phone,
                "ADDRESS" => $address,
                "APPROVERS" => $approvers,
                "BILLING_ADDRESS" => $addressBillingMandate["BILLING_ADDRESS"],
                "MANDATE" => $addressBillingMandate["MANDATE_ADDRESS"],
                "DELIVERY_ADDRESS" => $address,
                "NOTIFICATIONS_COUNT" => $user["NOTIFICATIONS_COUNT"]
            );
            
            return $array;
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Create and return GU Approvers array
     *
     * Receives an array with the GU approvers and create an array
     * with the size of the quantity informed. If the size of the approvers
     * array is lower than the quantity size, the remaining data of the array
     * will be created as empty strings.
     * 
     * @param array $approvers Array with GU approvers.
     * 
     * @return array
     * 
     * @access private
     * @author Grazielle Ferreira Antunes Brum
     */
    private function returnApproversGU($approvers)
    {
        $approversGU = [];
        
        foreach ($approvers as $approver) {
            array_push($approversGU, array(
                    "ID" => $approver["ID"],
                    "NAME" => $this->format->capitalizeWords($approver["NAME"])
                )
            );
        }

        return $approversGU;
    }
    
    /**
     * Create and return necessary data for GO User Info
     *
     * Receives the data needed to create the array with all the info
     * needed to return for the GO User Info process.
     *
     * @param Array $userInfo Array with user information.
     * 
     * @return array
     * @throws Exception
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function getUserInfoGO($userInfo)
    {
        try {
            if (isset($userInfo[0])) {
                $user = $userInfo[0];
                
                $address = $this->format->formatAddress($user["STREET"], $user["NEIGHBORHOOD"], $user["COUNTY"], $user["STATE"], $user["POSTAL_CODE"]);
                $name = $this->format->capitalizeWords($user["APPROVER_NAME"]);
                $phone = $this->format->formatPhone($user["PHONE"], true);
                $approvers = $this->returnApproversGO($user);
                
                $array = array(
                    "ADDRESS" => $address,
                    "APPROVERS" => $approvers,
                    "NAME" => $name,
                    "NOTIFICATIONS_COUNT" => $user["NOTIFICATIONS_COUNT"],
                    "PHONE" => $phone
                );
                
                return $array;
            } else {
                throw new \Exception("No data found!", 406);
            }
        } catch (\Exception $e) {
            throw $e;
        }
    }
    
    /**
     * Create and return GO Approvers array
     *
     * Receives an array with the GU approvers and create an array
     * containing the logged user data as main approver and a next approver
     * position if there is a next approver on the process.
     * 
     * @param array $approvers Array with GO approvers.
     * 
     * @return array
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function returnApproversGO($approvers)
    {
        $approversGO = array();
        
        array_push($approversGO, array(
                "ID" => $approvers["APPROVER_ID"], 
                "NAME" => $this->format->capitalizeWords($approvers["APPROVER_NAME"])
            )
        );
        
        if ($approvers["NEXT_APPROVER_ID"] != "" && $approvers["NEXT_APPROVER_ID"] != null) {
            array_push($approversGO, array(
                    "ID" => $approvers["NEXT_APPROVER_ID"], 
                    "NAME" => $this->format->capitalizeWords($approvers["NEXT_APPROVER_NAME"])
                )
            );
        }
        
        return $approversGO;
    }
}
