<?php

namespace SupplyTek\Utilities;

/**
 * Manipulates Files in Template Folder
 * 
 * Class to handle the templates inside the Templates Folder.
 * These Files are in general HTML templates of mails.
 * Also provides a function to replace the tags for the values.
 * 
 * @author Rômulo Alves Lousada
 */
class TemplateHandler
{
    private $directory;
    
    /**
     * Class constructor
     * 
     * Initialize an object's properties upon creation of the object.
     * 
     * @return void
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function __construct() 
    {
        $this->directory = realpath(__DIR__ . '/Templates');
    }
    
    /**
     * Get template of a file
     * 
     * Search for a file inside Template folder and return it's contents.
     * It can also replace some parts of the content before returning.
     *
     * @param array $file Array containing name and extension of the file.
     * @param array $search_replace (optional) Associative array with index/value to be replaced.
     *
     * @return array
     * @throws Exception
     * 
     * @access public
     * @author Rômulo Alves Lousada
     */
    public function getTemplate($file, $search_replace = array())
    {
        try {
            if ($file["name"]) {
                $path = $this->getFilePath($file);
                $content = $this->getFileContents($path);

                if (!empty($search_replace)) {
                    foreach ($search_replace as $index => $value) {
                        $search[] = $index;
                        $replace[] = $value;
                    }
                    
                    $content = $this->replaceTags($search, $replace, $content);
                }
                
                return $content;
            } else {
                throw new \Exception("Não foi enviado o nome do arquivo.", 400);
            }
        } catch (Exception $e) {
            throw new \Exception($e->getMessage(), $e->getCode());
        }
    }
    
    /**
     * Get file path
     * 
     * Creates the file path with data received
     *
     * @param array $file Array containing name and extension of the file.
     *
     * @return string
     * @throws Exception
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function getFilePath($file)
    {
        $fileName = $file["name"];
        $fileExtension = isset($file["extension"]) ? $file["extension"] : ".html";
        
        return $this->directory . '/' . $fileName . $fileExtension;
    }
    
    /**
     * Get file contents
     *
     * Check if the file path exists and return it's content.
     * 
     * @param string $path String with entire filepath to the file.
     *
     * @return string
     * @throws Exception
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function getFileContents($path)
    {
        if (file_exists($path)) {
            $content = file_get_contents($path);
            return mb_convert_encoding($content, 'UTF-8', mb_detect_encoding($content, 'UTF-8, ISO-8859-1', true));
        } else {
            throw new \Exception("Caminho de arquivo não encontrado.", 400);
        }
    }
    
    /**
     * Replace Tags
     *
     * Search for tags in the content of the file, and attempt to replace it
     * with the corresponding replace array.
     * If no modifications are done within this function, an Exception is
     * thrown to alert the developer that there must be errors in the arrays.
     * 
     * @param array $search Array with strings to be searched on the content.
     * @param array $replace Array with values to be replaced on the content.
     * @param string $content String with the file content.
     *
     * @return string
     * @throws Exception
     * 
     * @access private
     * @author Rômulo Alves Lousada
     */
    private function replaceTags($search, $replace, $content)
    {
        $content = str_replace($search, $replace, $content, $count);

        if ($count > 0) {
            return $content;
        } else {
            throw new \Exception("Nenhuma alteração das tags foi feita. Verifique os arrays enviados.", 400);
        }
    }
}